/*
  CPP_CONTEST=2013Cal-2011
  CPP_PROBLEM=B
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 2 mercurio 2
*/

/* RECORD
Serban, from Romania, Master student at the University of Murcia
December 2011
time 690
speed-up 26.84, 
The system does not accept to problem, but the result is correct in other systems.
There are some errors due to rounding.
It is included in the record table because the solution is totally different to
that of others provided
*/

/*RECORD in the new system
when the programs in the record table are run in the new system
formed by marte+mercurio (two hexacores),
it is the absolute record, with
time 1167
speed-up 10.88
In this system the solution was accepted
*/

#include <stdlib.h>
#include <omp.h>
#include <mpi.h>

typedef struct zero
{
	int beginning;
	int length;
} zero;

typedef struct line
{
	int size;
	double *elements;
	int nrZeroStructures;
	zero *zeros;
} line;

void define_MPI_datatype(MPI_Datatype *zeroType){
	zero z;
	int blocklen[2]={1,1}; //The first and last elements mark the beg and end of data structure
	MPI_Aint disp[2], origin;
	MPI_Datatype type[2]={MPI_INT, MPI_INT};
	int error;

	MPI_Address(&z, &origin ); 
	MPI_Address(&z.beginning, &(disp[0])); 
	MPI_Address(&z.length, &(disp[1])); 
	disp[0] -= origin;
	disp[1] -= origin;

	error=MPI_Type_create_struct(2, blocklen, disp, type, zeroType);
	MPI_Type_commit(zeroType);
}

void pack(int n, line *my_a, char *buffer, int bufferSize, MPI_Datatype zeroType, int *pos)
{
	int i;
	for(i=0;i<n;i++)
	{
		MPI_Pack(&my_a[i].size, 1, MPI_INT, buffer, bufferSize, pos, MPI_COMM_WORLD);
		MPI_Pack(my_a[i].elements, my_a[i].size, MPI_DOUBLE, buffer, bufferSize, pos, MPI_COMM_WORLD);
		MPI_Pack(&my_a[i].nrZeroStructures, 1, MPI_INT, buffer, bufferSize, pos, MPI_COMM_WORLD);
		MPI_Pack(my_a[i].zeros, my_a[i].nrZeroStructures, zeroType, buffer, bufferSize, pos, MPI_COMM_WORLD);
	}
}

void unpack(int n, line *my_a, char *buffer, int bufferSize, MPI_Datatype zeroType)
{
	int i, pos = 0;
	for(i=0;i<n;i++)
	{
		MPI_Unpack(buffer, bufferSize, &pos, &my_a[i].size, 1, MPI_INT, MPI_COMM_WORLD);
		my_a[i].elements = (double*)malloc(sizeof(double)*my_a[i].size);
		MPI_Unpack(buffer, bufferSize, &pos, my_a[i].elements, my_a[i].size, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Unpack(buffer, bufferSize, &pos, &my_a[i].nrZeroStructures, 1, MPI_INT, MPI_COMM_WORLD);
		my_a[i].zeros = (zero*)malloc(sizeof(zero)*my_a[i].nrZeroStructures);
		MPI_Unpack(buffer, bufferSize, &pos, my_a[i].zeros, my_a[i].nrZeroStructures, zeroType, MPI_COMM_WORLD);
	}
}

void processFirstMatrix(int n, double *m, line *lines)
{
	int i, j;
	
	for(i=0;i<n;i++)
	{
		lines[i].size = 0;
		lines[i].elements = (double*)malloc(sizeof(double)*n);
		lines[i].nrZeroStructures = 0;
		lines[i].zeros = (zero*)malloc(sizeof(zero)*n/2);
		for(j=0;j<n;j++)
		{
			if(m[i*n+j] == 0.)
			{
				int l=1;
				lines[i].zeros[lines[i].nrZeroStructures].beginning = j;
				j++;
				while(m[i*n+j] == 0. && j<n)
				{
					l++;
					j++;
				}
				j--;
				lines[i].zeros[lines[i].nrZeroStructures++].length = l;
			}
			else
			{
				lines[i].elements[lines[i].size++] = m[i*n+j];
			}
		}
	}
}

void processSecondMatrix(int n, double *m, line *lines)
{
	int i, j;
	for(j=0;j<n;j++)
	{
		lines[j].size = 0;
		lines[j].elements = (double*)malloc(sizeof(double)*n);
		lines[j].nrZeroStructures = 0;
		lines[j].zeros = (zero*)malloc(sizeof(zero)*n/2);
		for(i=0;i<n;i++)
		{
			if(m[i*n+j] == 0.)
			{
				int l=1;
				lines[j].zeros[lines[j].nrZeroStructures].beginning = i;
				i++;
				while(m[i*n+j] == 0. && i<n)
				{
					l++;
					i++;
				}
				i--;
				lines[j].zeros[lines[j].nrZeroStructures++].length = l;
			}
			else
			{
				lines[j].elements[lines[j].size++] = m[i*n+j];
			}
		}
	}
}

void reconstruct(line a, double *aux, int n)
{
	int k, z=0, real=0;
	for(k=0;k<n;k++)
	{
		if(z<=a.nrZeroStructures && a.zeros[z].beginning == k)
		{
			k+= a.zeros[z].length-1;
			z++;
		}
		else
		{
			aux[k] = a.elements[real];
			real++;
		}
	}
}

void multiply(line *a, line *b, double *my_c, int n, int colB)
{
	int id =0, total = 1, i ,j;
	#pragma omp parallel private(id, total)
	{ 
		total = omp_get_num_threads();
		id = omp_get_thread_num();
		#pragma omp for private(j) schedule(static)
		for (i = 0; i < n; i++) 
		{
			int k;
			double *aux = (double*)calloc(sizeof(double),n);
			reconstruct(a[i], aux, n);
			for(j=0;j<colB;j++)
			{
				double s = 0.;
				int z=0, real = 0;
				for(k=0;k<n;k++)
				{
					if(z<=b[j].nrZeroStructures && b[j].zeros[z].beginning == k)
					{
						k+= b[j].zeros[z].length-1;
						z++;
					}
					else
					{
						if(aux[k] != 0)
						{
							s += aux[k]*b[j].elements[real];
						}
						real++;
					}
				}
				my_c[i*colB+j] = s;
			}
		}
	}
}

void sec(int n,double *a,double *b,double *c,int node,int np)
{
        omp_set_num_threads(3);
	//inits
	MPI_Datatype zeroType;
	define_MPI_datatype(&zeroType);
	
	int colsPerProc, pos = 0;
	MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
	int bufferSize = sizeof(int)*n*2 + sizeof(double)*n*n + sizeof(zero)*n/2*n;
	char *buffer = (char*)malloc(bufferSize);
	int bBufferSize=sizeof(int)*n*2 + sizeof(double)*n*n + sizeof(zero)*n/2*n;
	char *bBuffer = (char*)malloc(bBufferSize);
	
	if(node != np-1)
		colsPerProc = n/np;
	else
		colsPerProc = n/np+n%np;
	MPI_Barrier(MPI_COMM_WORLD);
	//end inits
	
	
	
	//process and broadcast matrix A
	pos = 0;
	line *my_a = (line*)malloc(sizeof(line)*n);
	if(node == 0)
	{
		processFirstMatrix(n, a, my_a);
		pack(n, my_a, buffer, bufferSize, zeroType, &pos);
	}
	MPI_Bcast(&pos,1,MPI_INT,0,MPI_COMM_WORLD);
	MPI_Bcast(buffer,pos,MPI_PACKED,0,MPI_COMM_WORLD);
	if(node != 0)
		unpack(n, my_a, buffer, bufferSize, zeroType);
	//end
	
	
	
	//process and scatter matrix B
	pos = 0;
    int *displs = (int *)malloc(np*sizeof(int)); 
    int *scounts = (int *)malloc(np*sizeof(int)); 
	line *my_b = (line*)malloc(sizeof(line)*colsPerProc);
	if(node == 0)
	{
		int i, j;
		line *aux = (line*)malloc(sizeof(line)*n);
		processSecondMatrix(n, b, aux);
		pack(n, aux, buffer, bufferSize, zeroType, &pos);
		for (i=0; i<np; i++) 
		{ 
			if(i>0)
				displs[i] = displs[i-1]+scounts[i-1];
			else
				displs[i] = 0;
			scounts[i] = 0;
			for(j=i*colsPerProc;j<i*colsPerProc+colsPerProc;j++)
				scounts[i] += 2*sizeof(int)+aux[j].size*sizeof(double)+aux[j].nrZeroStructures*sizeof(zero);
		}
		i--;
		for(j=i*colsPerProc+colsPerProc;j<i*colsPerProc+colsPerProc+n%np;j++)
			scounts[i] += 2*sizeof(int)+aux[j].size*sizeof(double)+aux[j].nrZeroStructures*sizeof(zero);
	}
	MPI_Scatterv(buffer, scounts, displs, MPI_PACKED, bBuffer, bBufferSize, MPI_PACKED, 0, MPI_COMM_WORLD);
	unpack(colsPerProc, my_b, bBuffer, bBufferSize, zeroType);
	//end
	
	
	
	//multiplication
	double *my_c = (double*)malloc(sizeof(double)*n*colsPerProc);
	multiply(my_a, my_b, my_c, n, colsPerProc);
	//end
	
	//gather result
	MPI_Datatype acol, acoltype, bcol, bcoltype;
	if(node == 0)
	{
		MPI_Type_vector(n, 1, n, MPI_DOUBLE, &acol);       
		MPI_Type_commit(&acol);
		MPI_Type_create_resized(acol, 0, 1*sizeof(double), &acoltype);
		MPI_Type_commit(&acoltype);
		
		int i;
		for(i=0;i<np-1;i++)
		{
			displs[i]=i*colsPerProc;
			scounts[i] = colsPerProc;
		}
		displs[np-1]=(np-1)*colsPerProc;
		scounts[np-1] = n/np+n%np;
	}
	MPI_Type_vector(n, 1, colsPerProc, MPI_DOUBLE, &bcol);       
	MPI_Type_commit(&bcol);
	MPI_Type_create_resized(bcol, 0, 1*sizeof(double), &bcoltype);
	MPI_Type_commit(&bcoltype);
	MPI_Gatherv(my_c, colsPerProc, bcoltype, c, scounts, displs, acoltype, 0, MPI_COMM_WORLD);
	//end
}
